# =====================================
# generator=datazen
# version=3.1.4
# hash=7bf0794b4628951026a3aca53df62fb3
# =====================================
"""
instagram-bucketizer - Package's default entry-point.
"""

# built-in
import sys

# internal
from instagram_bucketizer.entry import main

if __name__ == "__main__":
    sys.exit(main(sys.argv))



"""
Bucket emojis

currently just looking at modifiable skin tone emojis and bucketing them
"""

# built-in
import argparse
from pathlib import Path

# third-party
from vcorelib.args import CommandFunction

# internal
from instagram_bucketizer.buckets.bucketizer import (
    Bucketizer,
    get_available_buckets,
)


def bucketize_cmd(args: argparse.Namespace):
    """
    Create and extend a bucket for each posts
    file given
    """

    bucketizer = Bucketizer(args.bucket)
    bucketizer.load_comments(args.comment_files[0])
    data = bucketizer.get_data()

    for comment in args.comment_files[1:]:
        next_bucket = Bucketizer(args.bucket)
        next_bucket.load_comments(comment)
        next_bucket.extend(data)

    bucketizer.write(data, args.output_file)


def add_bucketize_cmd(parser: argparse.ArgumentParser) -> CommandFunction:
    """Add arbiter-command arguments to its parser."""
    parser.add_argument(
        "comment_files", nargs="+", help="Comment files to parse", type=Path
    )
    parser.add_argument(
        "--output_file",
        "-o",
        type=Path,
        default=Path("output.xlsx"),
        help="Path to output comments loaded from a post",
    )
    parser.add_argument(
        "--bucket",
        "-b",
        help=f"Available: {get_available_buckets()}",
        required=True,
        type=str,
    )
    return bucketize_cmd



# =====================================
# generator=datazen
# version=3.1.4
# hash=8195916b986da0c0f4a634676fb6781b
# =====================================

"""
A module aggregating package commands.
"""

# built-in
from typing import List as _List
from typing import Tuple as _Tuple

# third-party
from vcorelib.args import CommandRegister as _CommandRegister

# internal
from instagram_bucketizer.commands.bucketize import add_bucketize_cmd
from instagram_bucketizer.commands.parse_post import add_parse_post_cmd
from instagram_bucketizer.commands.setup_session import add_setup_session_cmd


def commands() -> _List[_Tuple[str, str, _CommandRegister]]:
    """Get this package's commands."""

    return [
        (
            "bucketize",
            "bucketize comments",
            add_bucketize_cmd,
        ),
        (
            "parse_post",
            "parse a given post short code",
            add_parse_post_cmd,
        ),
        (
            "setup_session",
            "Used firefox cookies to setup session file",
            add_setup_session_cmd,
        ),
        ("noop", "command stub (does nothing)", lambda _: lambda _: 0),
    ]



"""
Base bucket abstract class
"""

# built-in
from abc import ABC, abstractmethod
from typing import Dict, List, Type, TypeVar, Union, cast

# third-party
import emoji as _emoji
from emoji import EMOJI_DATA, EmojiMatch

# generic type names dont need to be snake
# pylint: disable=invalid-name
TBucket = TypeVar("TBucket", bound="Bucket")
Comment = Dict[str, Union[int, str]]
Comments = List[Comment]
DataColumns = Dict[str, List[Union[int, str]]]

base_columns: DataColumns = {
    "post_id": [],
    "comment": [],
    "username": [],
    "datetime": [],
    "likes": [],
    "replies": [],
}

# Simplify the large structure to just english
EN_EMOJI_TO_UNICODE = {}
for uni, edat in EMOJI_DATA.items():
    EN_EMOJI_TO_UNICODE[edat["en"]] = uni


class Bucket(ABC):
    """
    Abstract class that should be able to be created from a comment
    process, then be called on to write to a dictionary
    of arrays to be exported to panda frame
    """

    def __init__(self, comment: Comment) -> None:
        self.comment = comment

    @staticmethod
    def get_columns() -> DataColumns:
        """
        Return this bucket's columns
        """
        return base_columns

    @staticmethod
    def process_emojis(comment_txt: str) -> List[str]:
        """
        Get a list of available emojis in a comment
        """
        tot_emoj = list(
            _emoji.analyze(
                _emoji.emojize(comment_txt), non_emoji=False, join_emoji=True
            )
        )
        emojis: List[str] = []
        if tot_emoj:
            for em in tot_emoj:
                assert isinstance(em.value, EmojiMatch)
                assert em.value.data
                emojis.append(em.value.data["en"])
        return emojis

    @classmethod
    def from_comment(cls: Type[TBucket], comment: Comment) -> TBucket:
        """
        Create child class based on given comment data
        """
        return cls(comment)

    def to_excel(self, data: DataColumns, post_id: str) -> None:
        """
        Add general comment data, then call child impl
        """

        data["post_id"].append(post_id)
        data["comment"].append(self.comment["text"])
        data["username"].append(self.comment["username"])
        data["datetime"].append(self.comment["created_at"])
        data["likes"].append(self.comment["likes_count"])
        data["replies"].append(
            len(cast(List[Comment], self.comment["answers"]))
        )
        self.to_excel_impl(data)

    @abstractmethod
    def to_excel_impl(self, data: DataColumns) -> None:
        """
        Given the premade data, adhereing to get bucket_cols for each bucket
        append this child classes data and return
        """



"""
    Bucket for categorizing emoji skin tone
    usage
"""

# built-in
from typing import Any, Dict, List, Type, cast

# internal
from instagram_bucketizer.buckets import (
    EN_EMOJI_TO_UNICODE,
    Bucket,
    Comment,
    DataColumns,
    TBucket,
    base_columns,
)

skin_tone_str = [
    "dark_skin_tone",
    "light_skin_tone",
    "medium_skin_tone",
    "medium-dark_skin_tone",
    "medium-light_skin_tone",
]


def get_tone_str(emoji: str):
    """
    Parse and return the skin tone
    substr if available
    """
    for tone in skin_tone_str:
        if "_" + tone in emoji:
            return tone
    return ""


def is_modifiable(emoji: str) -> bool:
    """
    Detect if the given emoji str
    is modifiable by skin tone
    """
    cleaned = emoji.replace(":", "")
    for tone in skin_tone_str:
        if tone in cleaned:
            return True

        if f":{cleaned}_{tone}:" in EN_EMOJI_TO_UNICODE:
            return True
    return False


def get_base_emoji(emoji: str) -> str:
    """
    Strip a tone substr if available
    """
    tone_str = get_tone_str(emoji)
    return emoji.replace(tone_str, "")


columns: Dict[str, List[Any]] = {
    "Emoji Total": [],
    "Emoji Modifiable": [],
    "Emoji Skin Neutral": [],
    "Emoji Skin Light": [],
    "Emoji Skin Medium Light": [],
    "Emoji Skin Medium": [],
    "Emoji Skin Medium Dark": [],
    "Emoji Skin Dark": [],
}


class SkinToneBucket(Bucket):
    """
    Bucket to add up all modifiable skin tone emojis
    """

    def __init__(self, comment: Comment):
        super().__init__(comment)
        self.comment_text = comment["text"]
        self.emojis = Bucket.process_emojis(cast(str, self.comment_text))

        self.total_emoji = len(self.emojis)
        self.tone_emoji_counts = {tone: 0 for tone in skin_tone_str}

        self.tone_emoji_counts["modifiable"] = 0
        self.tone_emoji_counts["neutral"] = 0
        for emoji in self.emojis:
            if is_modifiable(emoji):
                self.tone_emoji_counts["modifiable"] += 1
                if get_base_emoji(emoji) == emoji:
                    self.tone_emoji_counts["neutral"] += 1
                    continue

            tone_str = get_tone_str(emoji)
            if tone_str:
                self.tone_emoji_counts[tone_str] += 1

    @staticmethod
    def get_columns() -> DataColumns:
        """
        Get the data columns for this bucket
        """
        cols = columns
        cols.update(base_columns)
        return cols

    @classmethod
    def from_comment(cls: Type[TBucket], comment: Comment) -> TBucket:
        """
        Create child class based on given comment data
        """
        return cls(comment)

    def to_excel_impl(self, data: DataColumns) -> None:
        """
        Given the premade data, adhereing to get_bucket_cols for each bucket
        append this child classes data and return
        """

        data["Emoji Total"].append(self.total_emoji)
        data["Emoji Modifiable"].append(self.tone_emoji_counts["modifiable"])
        data["Emoji Skin Neutral"].append(self.tone_emoji_counts["neutral"])
        data["Emoji Skin Light"].append(
            self.tone_emoji_counts["light_skin_tone"]
        )
        data["Emoji Skin Medium Light"].append(
            self.tone_emoji_counts["medium-light_skin_tone"]
        )
        data["Emoji Skin Medium"].append(
            self.tone_emoji_counts["medium_skin_tone"]
        )
        data["Emoji Skin Medium Dark"].append(
            self.tone_emoji_counts["medium-dark_skin_tone"]
        )
        data["Emoji Skin Dark"].append(
            self.tone_emoji_counts["dark_skin_tone"]
        )



# =====================================
# generator=datazen
# version=3.1.4
# hash=647f4d4150952038b0f7695c6f3bbca4
# =====================================

"""
Useful defaults and other package metadata.
"""

DESCRIPTION = (
    "Small project to assist in research, categorizing instagram emoji usage"
)
PKG_NAME = "instagram-bucketizer"
VERSION = "0.1.0"



"""
    Pull comments from a post(s) and place into a comments file.
"""

# built-in
import argparse
from datetime import datetime
import json
from pathlib import Path
import time
from typing import Union

# third-party
import emoji
from instaloader import Instaloader, Post  # type: ignore [attr-defined]
from tqdm import tqdm
from vcorelib.args import CommandFunction
from vcorelib.paths import normalize

# internal
from instagram_bucketizer.buckets import Comments


def normalize_to_datestr(date: Union[str, int, datetime]) -> str:
    """
    Given a epoch, str, or datetime, make sure its a str
    """

    if isinstance(date, str):
        return date
    if isinstance(date, int):
        return str(datetime.fromtimestamp(date))
    if isinstance(date, datetime):
        return str(date)
    assert False, f"Given data {date} is invalid! {type(date)}"


def normalize_to_epoch(date: Union[str, float, int, datetime]) -> float:
    """
    Given a epoch, str, or datetime obj make sure its an int
    """

    if isinstance(date, str):
        utc_time = datetime.strptime(date, "%Y-%m-%d %H:%M:%S")
        return ((utc_time) - datetime(1970, 1, 1)).total_seconds()
    if isinstance(date, int):
        return float(date)
    if isinstance(date, datetime):
        return date.timestamp()
    if isinstance(date, float):
        return date
    assert False, f"Given data {date} is invalid! {type(date)}"


def to_emoji_str(emoji_str: str) -> str:
    """
    Translate utf-16 to emoji strings
    """
    return emoji_str.encode("utf-16-be", "surrogatepass").decode("utf-16-be")


def write_post(
    args: argparse.Namespace, post_code: str, comments: Comments
) -> None:
    """
    Write data out
    """
    output_dir = normalize(args.output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)
    sorted_data = sorted(
        comments, key=lambda x: normalize_to_epoch(x["created_at"])
    )
    for data in sorted_data:
        data["text"] = emoji.demojize(to_emoji_str(str(data["text"])))

    full_out = {"post_code": post_code, "comments": sorted_data}
    with open(
        output_dir.joinpath(f"{args.prefix}_{post_code}.json"),
        "w",
        encoding="UTF-8",
    ) as com_file:
        com_file.write(json.dumps(full_out, indent=4))


def parse_post_cmd(args: argparse.Namespace):
    """
    for each post code, get data, load into dictionary and return
    """
    assert args.username, "No username provided!"
    loader = Instaloader()
    loader.load_session_from_file(args.username)

    for post_code in args.post_codes:
        post = Post.from_shortcode(loader.context, post_code)
        print(f"Loading comments from {post_code}")
        with loader.context.error_catcher("Comment Parsing:"):
            ret_data: Comments = []
            with tqdm(total=post.comments) as progress_counter:
                for comment in post.get_comments():
                    ret_data.append(
                        {
                            "id": comment.id,
                            "created_at": normalize_to_datestr(
                                comment.created_at_utc
                            ),
                            "text": comment.text,
                            "username": comment.owner.username,
                            "likes_count": comment.likes_count,
                            "answers": sorted(  # type: ignore[dict-item]
                                [
                                    {
                                        "id": answer.id,
                                        "created_at": normalize_to_datestr(
                                            answer.created_at_utc
                                        ),
                                        "text": answer.text,
                                        "username": answer.owner.username,
                                        "likes_count": answer.likes_count,
                                    }
                                    for answer in comment.answers
                                ],
                                key=lambda x: normalize_to_epoch(
                                    x["created_at"]
                                ),
                            ),
                        }
                    )
                    progress_counter.update(1)
            write_post(args, post_code, ret_data)

        if post_code != args.post_codes[-1]:
            for _ in tqdm(range(args.wait)):
                time.sleep(1)


def add_parse_post_cmd(parser: argparse.ArgumentParser) -> CommandFunction:
    """Add arbiter-command arguments to its parser."""
    parser.add_argument(
        "post_codes", nargs="+", help="Post codes to parse", type=str
    )
    parser.add_argument("-u", "--username", type=str)
    parser.add_argument(
        "--output_dir",
        type=Path,
        default=Path("."),
        help="Path to output comments from posts",
    )
    parser.add_argument(
        "-P", "--prefix", help="Output file prefix", default=""
    )
    parser.add_argument(
        "-w",
        "--wait",
        type=int,
        default=60,
        help="Place a wait between parsing posts",
    )
    return parse_post_cmd



"""
    Setup a session file based on web browser cookies
"""
# This command is a workaround for instaloader
# and thus contains some improper accesses etc..
# see https://instaloader.github.io/troubleshooting.html
# mypy: ignore-errors

# built-in
import argparse
from glob import glob
from os.path import expanduser
from platform import system
from sqlite3 import OperationalError, connect

from instaloader import ConnectionException, Instaloader

# third-party
from vcorelib.args import CommandFunction

WINDOWS_PATH = (
    "~/AppData/Roaming/" + "Mozilla/Firefox/Profiles/*/cookies.sqlite"
)
DARWIN_PATH = (
    "~/Library/Application Support/" + "Firefox/Profiles/*/cookies.sqlite"
)
LINUX_PATH = "~/.mozilla/firefox/*/" + "cookies.sqlite"


def get_cookiefile():
    """
    Find a cookie file
    """

    default_cookiefile = {
        "Windows": WINDOWS_PATH,
        "Darwin": DARWIN_PATH,
    }.get(system(), LINUX_PATH)
    cookiefiles = glob(expanduser(default_cookiefile))
    if not cookiefiles:
        raise SystemExit(
            "No Firefox cookies.sqlite file found. Use -c COOKIEFILE."
        )
    return cookiefiles[0]


def import_session(cookiefile, sessionfile):
    """
    Save a session file
    """

    print(f"Using cookies from {cookiefile}.")
    conn = connect(f"file:{cookiefile}?immutable=1", uri=True)
    try:
        cookie_data = conn.execute(
            "SELECT name, value FROM moz_cookies"
            + " WHERE baseDomain='instagram.com'"
        )
    except OperationalError:
        cookie_data = conn.execute(
            "SELECT name, value FROM moz_cookies"
            + " WHERE host LIKE '%instagram.com'"
        )
    instaloader = Instaloader(max_connection_attempts=1)

    # this session setup is generally a workaround
    # pylint: disable=protected-access
    instaloader.context._session.cookies.update(cookie_data)
    username = instaloader.test_login()
    if not username:
        raise SystemExit(
            "Not logged in. Are you logged in successfully in Firefox?"
        )
    print(f"Imported session cookie for {username}.")
    instaloader.context.username = username
    instaloader.save_session_to_file(sessionfile)


def setup_session_cmd(args: argparse.Namespace):
    """
    Setup a session file command
    """

    try:
        import_session(args.cookiefile or get_cookiefile(), args.sessionfile)
    except (ConnectionException, OperationalError) as e:
        raise SystemExit(f"Cookie import failed: {e}") from e


def add_setup_session_cmd(parser: argparse.ArgumentParser) -> CommandFunction:
    """Add arbiter-command arguments to its parser."""
    parser.add_argument("-c", "--cookiefile")
    parser.add_argument("-f", "--sessionfile")
    return setup_session_cmd



"""
Bucketizer class
"""

# built-in
import json
from typing import List

import pandas as pd

# third party
from vcorelib.paths import Pathlike, normalize

# internal
from instagram_bucketizer.buckets import Bucket, DataColumns

# This is used in detecting bucket subclasses
# pylint: disable=unused-import
# flake8: noqa
from instagram_bucketizer.buckets.emoji_tone import SkinToneBucket


def get_available_buckets() -> List[str]:
    """
    Return a list of available buckets
    """
    return [cls.__name__ for cls in Bucket.__subclasses__()]


def get_bucket(name: str) -> type[Bucket]:
    """
    Get a bucket class from string
    """
    for cls in Bucket.__subclasses__():
        if cls.__name__ == name:
            return cls
    assert False, (
        f"Bucket {name} is not" + f" available ({get_available_buckets()})"
    )


class Bucketizer:
    """
    Given a string, get corresponding bucket child name
    hand list of comment dicts, create buckets
    print to excel
    """

    def __init__(self, bucket: str) -> None:
        available = get_available_buckets()
        assert (
            bucket in available
        ), f"Bucket {bucket} not in availble bucket!\n available: {available}!"

        self.bucket: type[Bucket] = get_bucket(bucket)
        self.comments: List[Bucket] = []
        self.post_code = ""

    def load_comments(self, comments_file: Pathlike) -> None:
        """
        Load all the comments in a file
        """
        comments_file = normalize(comments_file)
        with open(comments_file, "r", encoding="utf-8") as read_fd:
            data = json.load(read_fd)
            self.post_code = data["post_code"]
            for comment in data["comments"]:
                self.comments.append(self.bucket.from_comment(comment))

    def extend(self, data: DataColumns) -> None:
        """
        Given data struct, extend it with own data
        """
        for comment in self.comments:
            comment.to_excel(data, self.post_code)

    def get_data(self) -> DataColumns:
        """
        Return a header row plus bucket data
        """
        data = self.bucket.get_columns()
        self.extend(data)
        return data

    def write(self, data: DataColumns, output_file: Pathlike) -> None:
        """
        Write given data to the output file
        """
        output_file = normalize(output_file)
        output_data = pd.DataFrame(data=data)
        output_data.to_excel(output_file)
